# Generate Delphi wrapper for HDF5 library.
# by Andrey Paramonov

import sys
import os.path
import argparse
import networkx as nx
import datetime
import re

parser = argparse.ArgumentParser(description = 'Generate Delphi wrapper for HDF5 library.')
parser.add_argument('srcdir', help = 'directory containing HDF5 *.h files.',
                    nargs = '?', default = '.')
args = parser.parse_args()

def parsedeps(header, graph):
    if header.startswith('H5') and header not in graph.onodes:
        graph.onodes.append(header)
    for line in open(os.path.join(args.srcdir, header)):
        m = re.match('#include "(H5.*public.h)".*', line)
        if m:
            include = m.group(1)
            if header.startswith('H5'):
                if include not in graph.onodes:
                    graph.onodes.append(include)
                graph.add_edge(header, include)
            parsedeps(include, graph)

defs = ''
classname = 'THDF5Dll'
cfields = ''
fields = ''
methods = ''
props = ''
init = ''
cinit = ''
impl = ''

template = \
'''unit hdf5dll;

// Delphi wrapper for HDF5 library.

// Auto-generated {date} by hdf5pas.py.

interface

uses
  windows;

type
  UInt64 = Int64;
  PUInt64 = ^Int64;
  PCardinal = ^Cardinal;
  PPointer = ^Pointer;
  PPAnsiChar = ^PAnsiChar;

type
  int32_t = Integer;
  Pint32_t = ^int32_t;
  uint32_t = Cardinal;
  Puint32_t = ^uint32_t;
  int64_t = Int64;
  Pint64_t = ^int64_t;
  uint64_t = UInt64;
  Puint64_t = ^uint64_t;
  time_t = Integer;
  Ptime_t = ^time_t;
  size_t = Cardinal;
  Psize_t = ^size_t;
  ssize_t = Integer;
  Pssize_t = ^ssize_t;
  off_t = Integer;
  Poff_t = ^off_t;
  PFILE = Pointer;

type
  hsize_t = UInt64;
  Phsize_t = ^hsize_t;
  hssize_t = Int64;
  Phssize_t = ^hssize_t;
  haddr_t = UInt64;
  Phaddr_t = ^haddr_t;
const
  HADDR_UNDEF = -1;

{defs}

type
  {classname} = class
  private
    FHandle: THandle;

{fields}
{cfields}

  public
    constructor Create(APath: string);
    destructor Destroy; override;

{methods}
{props}

    property Handle: THandle read FHandle;
    function IsValid: Boolean;
  end;

implementation

{{ {classname} }}
constructor {classname}.Create(APath: string);

  function GetDllProc(AModule: THandle; AName: string): Pointer;
  begin
    Result := GetProcAddress(AModule, PChar(AName));
    Assert(Assigned(Result));
  end;

begin
  inherited Create;
  FHandle := LoadLibrary(PChar(APath));

{init}

  H5open;
{cinit}
end;

destructor {classname}.Destroy;
begin
  if FHandle <> 0 then
    FreeLibrary(FHandle);
  inherited;
end;

function {classname}.IsValid: Boolean;
begin
  Result := (FHandle <> 0);
end;

{impl}

end.
'''

def parse(header):

    def smartjoin(sep, *args):
        if args[0]:
            return sep.join(args)
        else:
            return args[1]

    def stripcomment(s):
        return re.sub(' *(\(\*.*\*\))?$', '', s)

    def strtoint(value):
        value = re.sub('^\(\(.*\)\)$', r'\1', value.strip())
        if value.startswith('('):
            tokens = re.findall('(\((.*?)\)( *|$))|([^()]+$)', value)
            value = (tokens[-1][1] or tokens[-1][3]).strip()
        else:
            tokens = None
        value = value.rstrip('uL')
        try:
            result = int(value, 0)
            if tokens:
                for token in reversed(tokens[:-1]):
                    typ = token[1].strip()
                    (name, typ) = convnametype('', typ)
                    result = '{}({})'.format(typ, result)
        except ValueError:
            m = re.match('(.*) << (.*)', value)
            if m:
                result = '{} shl {}'.format(m.group(1), int(m.group(2), 0))
            else:
                return
        return result

    def strtofloat(value):
        try:
            value = value.rstrip('f')
            return float(value)
        except ValueError:
            pass

    def convnametype(cname, ctype):
        # Convert C-stype variable/constant/field declaration to Delphi-style

        def replace(where, olditems, newitem):
            items = where
            for item in olditems:
                if item in items:
                    items = [s for s in items if s != item]
                else:
                    return where
            return items + [newitem]

        typ = ctype.replace('*', ' * ')
        while '  ' in typ:
            typ = typ.replace('  ', ' ')
        typ = typ.strip().split(' ')
        stars = len([s for s in cname if s == '*'])
        name = cname.strip('* ')
        typ += ['*']*stars
        if name.endswith('[]'):
            name = name.rstrip('[]')
            typ += ['*']

        if name == 'type':
            name = 'typ'
        elif name == 'object':
            name = 'obj'
        elif name == 'end':
            name = 'end_'
        elif name == 'file':
            name = 'file_'

        typ = [s for s in typ if s != 'const']
        typ = replace(typ, ['unsigned', 'long', 'long', 'int'], 'UInt64')
        typ = replace(typ, ['unsigned', 'long', 'long'], 'UInt64')
        typ = replace(typ, ['long', 'long', 'int'], 'Int64')
        typ = replace(typ, ['long', 'long'], 'Int64')
        typ = replace(typ, ['unsigned', 'long', 'int'], 'Cardinal')
        typ = replace(typ, ['unsigned', 'long'], 'Cardinal')
        typ = replace(typ, ['long', 'int'], 'Integer')
        typ = replace(typ, ['long'], 'Integer')
        typ = replace(typ, ['unsigned', 'short', 'int'], 'Word')
        typ = replace(typ, ['unsigned', 'short'], 'Word')
        typ = replace(typ, ['short', 'int'], 'ShortInt')
        typ = replace(typ, ['short'], 'ShortInt')
        typ = replace(typ, ['unsigned', 'int'], 'Cardinal')
        typ = replace(typ, ['int'], 'Integer')
        typ = replace(typ, ['unsigned', 'char'], 'Byte')
        typ = replace(typ, ['char'], 'AnsiChar')
        typ = replace(typ, ['unsigned'], 'Cardinal')
        if '*' in typ:
            typ = replace(typ, ['void'], 'ointer')

        stars = len([s for s in typ if s == '*'])
        typ = 'P'*stars + ''.join([s for s in typ if s != '*'])
        return (name, typ)

    def preprocess(lines):
        '''
        Parse and strip off pre-processor directives.
        Currently all #if/#ifdef/#ifndef are considered as false.
        '''

        print >> sys.stderr,\
            '{}: Pre-processing...'.format(header)
        ifdef = 0
        result = []
        for line in lines:
            line = line.strip('\n').expandtabs()
            if line.strip() == '':
                line = ''
            line = line.replace('/*', '(*').replace('*/', '*)')
            hdef = '_{}_H'.format(os.path.splitext(header)[0])
            if re.match('#ifndef {}'.format(hdef), line) or \
               re.match('#define {}'.format(hdef), line):
                pass
            elif line.startswith('#if') or \
                 line.startswith('#ifdef') or \
                 line.startswith('#ifndef'):
                ifdef += 1
            elif line.startswith('#endif'):
                ifdef -= 1
            elif not ifdef:
                if line.startswith('#include') or line.startswith('#undef'):
                    pass
                else:
                    result.append(line)
        print >> sys.stderr,\
            '{}: {} of {} lines left'.format(header, len(result), len(lines))
        return result

    lines = open(os.path.join(args.srcdir, header)).readlines()
    lines = preprocess(lines)

    print >> sys.stderr,\
        '{}: Parsing...'.format(header)

    def process(state, stateinfo, comment):

        def procdefine(lines):
            '''
            Process sequence of #define's.
            '''

            global props

            result = ''
            comment = False
            for line in lines.split('\n'):
                m = re.match(' *?(((\(\*)|( \*)).*)', line)
                if m:
                    comment = True
                    if len(result) > 0:
                        result += '\n' + m.group(1)
                else:
                    m = re.match(r'#define +(.*?) +([^\\]+)$', stripcomment(line))
                    if m:
                        comment = False
                        mm = re.search('\(\*(.*)\*\)', line.strip())
                        comment = mm.group(1) if mm else None
                        (name, value) = m.groups()
                        value = re.sub('^\((.*)\)$', r'\1', value)
                        value = re.sub('^H5CHECK ', '', value)
                        value = value.replace('sizeof', 'SizeOf')
                        comment = ' '.join(['(*', comment.strip(), '*)'] if comment else '')
                        if '?' in value or ',' in value:
                            print >> sys.stderr,\
                                'WARN: {}'.format(line)
                        elif value.startswith('H5OPEN'):
                            props += '    property {}: hid_t read F{};\n'.format(name, value.split(' ')[-1].strip('_g'))
                        elif 'SIZEOF' in name:
                            pass
                        elif strtoint(value) != None:
                            result += '\n  {} = {};  {}'.format(name, strtoint(value), comment)
                        elif strtofloat(value) != None:
                            result += '\n  {} = {};  {}'.format(name, strtofloat(value), comment)
                        elif value.startswith('"') and value.endswith('"'):
                            result += "\n  {} = '{}';  {}".format(name, value.strip('"'), comment)
                        elif len(value.split('|')) > 1:
                            result += '\n  {} = {};  {}'.format(name,
                                                                ' or '.join([item.strip()
                                                                             for item in value.split('|')]),
                                                                comment)
                        elif name.startswith('H5T_INTEL') or \
                             name.startswith('H5T_ALPHA') or \
                             name.startswith('H5T_MIPS'):
                            props += '    property {}: hid_t read F{};\n'.format(name, value)
                        else:
                            result += '\n  {} = {};  {}'.format(name, value, comment)
                    elif comment:
                        result += '\n' + line
                    else:
                        print >> sys.stderr,\
                            'WARN: {}'.format(line)
            return result

        def proctypedef(lines):
            '''
            Process sequence of typedefs.
            '''

            def process(prevstate, state, stateinfo):
                '''
                Process one typedef.
                '''

                result = ''
                commentstart = None

                if len(stateinfo) == 1:
                    if state == 'enum':
                        stateinfo[0] = stateinfo[0].replace('typedef enum', 'typedef')
                    elif state == 'struct':
                        stateinfo[0] = stateinfo[0].replace('typedef struct', 'typedef')
                    state = 'other'

                if state == 'enum':
                    '''
                    Enumerated type declaration.
                    '''

                    result += '\ntype'
                    name = stateinfo[-1].strip('}; ') or stateinfo[0].split(' ')[2]
                    result += '\n  {} = Integer;'.format(name)
                    result += '\n  P{name} = ^{name};'.format(name = name)
                    result += '\nconst'
                    maxval = -1
                    for line in stateinfo[1:len(stateinfo) - 1]:
                        if stripcomment(line).strip() == '{':
                            continue
                        m = re.match(' *([^ *(),]+)( *= ?([^,]+))?,?', stripcomment(line))
                        if m:
                            (name, dummy, value) = m.groups()
                            value = strtoint(value) if value else None
                            mm = re.search('\(\*(.*)\*\)', line.strip())
                            comment = mm.group(1) if mm else None
                            comment = ' '.join(['(*', comment.strip(), '*)'] if comment else '')
                            if value != None:
                                maxval = value
                            else:
                                maxval += 1
                                value = maxval
                            result += '\n  {} = {};  {}'.format(name, value, comment)
                            if not comment:
                                m = re.search(' (\(\*.*)', line)
                                if m:
                                    commentstart = m.group(1)
                        elif not stripcomment(line).strip():
                            result += '\n' + line.strip()
                        elif re.match(' *(\(\*.*)', line):
                                m = re.search(' (\(\*.*)', line)
                                if m:
                                    commentstart = m.group(1)
                        elif re.match(' *( \*.*)', line):
                            if commentstart:
                                result += '\n' + commentstart
                                commentstart = None
                            result += '\n' + re.sub(' *( \*.*)', r'\1', line)
                        else:
                            print >> sys.stderr,\
                                'WARN: {}'.format(line)

                elif state == 'struct':
                    '''
                    Compound type (struct) declaration.
                    '''
                    # TODO: Struct declarations are not converted
                    # (how to support byte alignment?).

                    result += '\ntype'
                    name = stateinfo[-1].strip('}; ') or stateinfo[0].split(' ')[2]
                    result += '\n  PP{name} = ^P{name};'.format(name = name)
                    result += '\n  P{} = Pointer;'.format(name)

                elif state == 'other':
                    comments = None
                    for i in range(len(stateinfo)):
                        if stateinfo[i].startswith('(*'):
                            comments = stateinfo[i:]
                            stateinfo = stateinfo[:i]
                            break
                    if len(stateinfo) == 1 and re.match('typedef *([^(),]*) +([^(),]*);', stateinfo[0]):
                        '''
                        Type synonym.
                        '''

                        (typ, name) = re.match('typedef *([^(),]*) +([^(),]*);', stateinfo[0]).groups()
                        (name, typ) = convnametype(name.strip(), typ.strip())
                        if name != typ:
                            if prevstate != 'other':
                                result += 'type\n'
                            if name.endswith(']'):
                                result += '  P{} = P{};'.format(re.sub('\[.*', '', name), typ)
                            else:
                                result += '  {} = {};'.format(name, typ)
                                result += '\n  P{name} = ^{name};'.format(name = name)
                    else:
                        '''
                        Procedural type declaration.
                        '''

                        if prevstate != 'other':
                            result += 'type\n'
                        signature = re.sub('\(\*[^()]*?\*\)', '', ' '.join(stateinfo))
                        while '  ' in signature:
                            signature = signature.replace('  ', ' ')
                        (rettype, name, args) = re.match('typedef (.*) ?\(\*(.*)\) ?\((.*)\);', signature).groups()
                        args = [s.strip() for s in args.split(',')]
                        for i in range(len(args)):
                            arg = args[i].split(' ')
                            if len(arg) < 2:
                                arg.append('p')
                            atyp = ' '.join(arg[:-1])
                            aname = arg[-1]
                            (aname, atyp) = convnametype(aname, atyp)
                            args[i] = '{}: {}'.format(aname, atyp)
                        rettype = convnametype('', rettype)[-1]
                        if rettype == 'void':
                            result += '  {} = procedure({}); cdecl;'.format(name, '; '.join(args))
                        else:
                            result += '  {} = function({}): {}; cdecl;'.format(name, '; '.join(args), rettype)
                        result += '\n  P{name} = ^{name};'.format(name = name)
                    if comments:
                        result += '\n'.join([''] + comments)
                return result

            result = ''
            prevstate = None
            state = None
            stateinfo = []
            for line in lines.split('\n'):
                line = re.sub('^enum', 'typedef enum', line)
                line = re.sub('^struct', 'typedef struct', line)
                if line.startswith('typedef enum'):
                    result += '\n' + process(prevstate, state, stateinfo)
                    prevstate = state
                    state = 'enum'
                    stateinfo = []
                elif line.startswith('typedef struct'):
                    result += '\n' + process(prevstate, state, stateinfo)
                    prevstate = state
                    state = 'struct'
                    stateinfo = []
                elif line.startswith('typedef '):
                    result += '\n' + process(prevstate, state, stateinfo)
                    prevstate = state
                    state = 'other'
                    stateinfo = []
                if state:
                    stateinfo.append(line)
                else:
                    print >> sys.stderr,\
                        'WARN: {}'.format(line)
            if state:
                result += '\n' + process(prevstate, state, stateinfo)
            return result

        def procexport(lines):
            '''
            Process sequence of exported symbols.
            '''

            global defs, fields, cfields, methods, props, init, cinit, impl

            signature = None
            for line in lines.split('\n'):
                line = re.sub('[(/]\*.*?\*[)/]', '', line.strip())
                if line.startswith('H5_DLLVAR'):
                    '''
                    Exported variable.
                    '''

                    (dummy, ctype, cname) = line.split(' ')
                    cname = cname.strip('_g;')
                    (cname, ctype) = convnametype(cname, ctype)
                    cfields += '    F{}: {};\n'.format(cname, ctype)
                    cinit += "  F{cname} := P{ctype}(GetDllProc(FHandle, '{cname}_g'))^;\n".format(cname = cname, ctype = ctype)

                else:
                    '''
                    Exported procedure.
                    '''

                    signature = smartjoin(' ', signature, line)
                    if not ')' in line:
                        continue

                    signature = re.sub('\(\*.*?\*\)', '', signature)
                    signature = signature.replace(' (', '(').\
                                replace('*', ' * ').replace('  ', ' ')

                    (rettype, fname, args) = re.match('H5_DLL (.*) (.*) ?\((.*)\);', signature).groups()
                    if args != 'void':
                        args = [s.strip() for s in args.split(',')]
                    else:
                        args = []
                    varargs = False
                    for i in range(len(args)):
                        arg = args[i].split(' ')
                        if len(arg) < 2 and args[i] != '...':
                            arg.append('p')
                        atyp = ' '.join(arg[:-1])
                        aname = arg[-1]
                        (aname, atyp) = convnametype(aname, atyp)
                        if args[i] != '...':
                            args[i] = '{}: {}'.format(aname, atyp)
                        else:
                            args[i] = None
                            varargs = True
                    args = [s for s in args if s]
                    rettype = convnametype('', rettype)[-1]

                    if len(args) > 0:
                        fdef = '(' + '; '.join(args) + ')'
                    else:
                        fdef = ''
                    fdef = fdef + ': ' + rettype

                    if varargs:
                        fields += '    // F{}: function{}; cdecl; varargs;\n'.format(fname, fdef)
                        methods += '    // function {}{}; cdecl; varargs;\n'.format(fname, fdef)
                        print >> sys.stderr,\
                            'ERROR: Ignoring varargs procedure {}.'.format(fname)
                    else:
                        fields += '    F{}: function{}; cdecl;\n'.format(fname, fdef)
                        methods += '    function {}{};\n'.format(fname, fdef)
                        init += "  @F{0} := GetDllProc(FHandle, '{0}');\n".format(fname)
                        impl += '\nfunction {name}.{fname}{fdef};\nbegin\n  Result := F{fname}({argnames});\nend;\n'.format(
                            name = classname, fname = fname, fdef = fdef,
                            argnames = ', '.join(arg.split(':')[0] for arg in args))
                    signature = None

        global defs, fields, cfields, methods, props, init, cinit, impl
        if stateinfo:
            stateinfo = stateinfo.strip('\n')
        if state == 'define':
            newdefs = procdefine(stateinfo).lstrip('\n')
            if len(newdefs) > 0:
                if comment:
                    defs += '\n'
                    defs += comment.strip('\n') + '\n'
                defs += 'const\n'
                defs += newdefs
                defs += '\n'
        elif state == 'typedef':
            newdefs = proctypedef(stateinfo).lstrip('\n')
            if len(newdefs) > 0:
                if comment:
                    defs += '\n'
                    defs += comment.strip('\n') + '\n'
                defs += newdefs
                defs += '\n'
        elif state == 'export':
            newdefs = procexport(stateinfo)

    global state, stateinfo, comment
    state = None
    stateinfo = None
    comment = None

    def setstate(newstate):
        global state, stateinfo, comment
        if stateinfo and stateinfo.endswith('\n'):
            if state:
                process(state, stateinfo, comment)
            state = newstate
            stateinfo = None
            comment = None
        elif newstate != state:
            if newstate == 'comment':
                if state:
                    return
            else:
                if state == 'comment':
                    comment = stateinfo
            process(state, stateinfo, comment)
            if state != 'comment':
                comment = None
            state = newstate
            stateinfo = None

    for line in lines + ['']:
        if line.startswith('(*'):
            setstate('comment')
        elif not state and line.lstrip(' ').startswith('(*'):
            setstate('comment')
        elif line.startswith('#define'):
            setstate('define')
        elif line.startswith('typedef') or \
             line.startswith('struct') or \
             line.startswith('enum'):
            setstate('typedef')
        elif line.startswith('H5_DLL'):
            setstate('export')
        elif line and not state:
            raise Exception(header, line)

        if state:
            stateinfo = smartjoin('\n', stateinfo, line)
    setstate(None)

    print >> sys.stderr

graph = nx.DiGraph()
graph.onodes = []
parsedeps('hdf5.h', graph)
paths = nx.all_pairs_shortest_path_length(graph)
for header in sorted(graph.onodes, key = lambda header: len(paths[header])):
    parse(header)

for line in template.format(date = datetime.date.today(),
                            defs = defs.strip('\n'),
                            classname = classname,
                            fields = fields.strip('\n'),
                            cfields = cfields.strip('\n'),
                            methods = methods.strip('\n'),
                            props = props.strip('\n'),
                            init = init.strip('\n'),
                            cinit = cinit.strip('\n'),
                            impl = impl.strip('\n')).split('\n'):
    print line.rstrip()
